Index: enum.c
===================================================================
RCS file: /usr/cvsroot/asterisk/enum.c,v
retrieving revision 1.25
diff -u -r1.25 enum.c
--- enum.c	21 Apr 2005 06:02:43 -0000	1.25
+++ enum.c	2 May 2005 22:56:44 -0000
@@ -7,6 +7,9 @@
  *
  * Funding provided by nic.at
  *
+ * Proper handling of order and priority with multiple NAPTR records
+ * written by Edwin Groothuis <edwin@mavetju.org>
+ *
  * Distributed under the terms of the GNU GPL
  *
  */
@@ -56,6 +59,7 @@
 static int enumver = 0;
 
 AST_MUTEX_DEFINE_STATIC(enumlock);
+AST_MUTEX_DEFINE_STATIC(enumcachelock);
 
 struct naptr {
 	unsigned short order;
@@ -81,7 +85,7 @@
 }
 
 /*--- parse_naptr: Parse DNS NAPTR record used in ENUM ---*/
-static int parse_naptr(unsigned char *dst, int dstsize, char *tech, int techsize, unsigned char *answer, int len, char *naptrinput)
+static int parse_naptr(unsigned char *dst, int dstsize, char *tech, int techsize, unsigned char *answer, int len, unsigned short *pref, unsigned short *order,char *naptrinput)
 {
 	unsigned char *oanswer = answer;
 	unsigned char flags[80] = "";
@@ -97,7 +101,7 @@
 	int d_len = sizeof(temp) - 1;
 	regex_t preg;
 	regmatch_t pmatch[9];
-
+	struct naptr *naptrhdr;
 
 	dst[0] = '\0';
 	
@@ -105,6 +109,9 @@
 		ast_log(LOG_WARNING, "NAPTR record length too short\n");
 		return -1;
 	}
+	naptrhdr=(struct naptr *)answer;
+	*pref=ntohs(naptrhdr->pref);
+	*order=ntohs(naptrhdr->order);
 	answer += sizeof(struct naptr);
 	len -= sizeof(struct naptr);
 	if ((res = parse_ie(flags, sizeof(flags) - 1, answer, len)) < 0) {
@@ -159,11 +166,14 @@
 	} else if ((!strncasecmp(services, "e2u+tel", 7)) || 
 	    (!strncasecmp(services, "tel+e2u", 7))) {
 		strncpy(tech, "tel", techsize -1); 
+	} else if ((!strncasecmp(services, "e2u+mailto", 10)) || 
+	    (!strncasecmp(services, "mailto+e2u", 7))) {
+		strncpy(tech, "mailto", techsize -1); 
 	} else if (!strncasecmp(services, "e2u+voice:", 10)) {
 		strncpy(tech, services+10, techsize -1); 
 	} else {
 		ast_log(LOG_DEBUG, 
-		"Services must be e2u+${tech}, ${tech}+e2u, or e2u+voice: where $tech is from (sip, h323, tel, iax, iax2). \n");
+		"Services must be e2u+${tech}, ${tech}+e2u, or e2u+voice: where $tech is from (sip, h323, tel, iax, iax2, mailto). Got service '%s'\n",services);
 		return 0;
 	}
 
@@ -187,7 +197,7 @@
 
 	pattern = regexp + 1;
 	*delim2 = 0;
-	subst   = delim2 + 1;
+	subst	= delim2 + 1;
 	regexp[regexp_len-1] = 0;
 
 #if 0
@@ -245,6 +255,29 @@
 	return 0;
 }
 
+/*
+ * The enum_cache is a linked list with channel_uniqueids, birthtimes of
+ * the record and the NAPTR record information.
+ * The birthtime is so that we know when to throw it away.
+ * The channel_uniqueid is to know which channel the records are ssociated with
+ * And the information from the NAPTR record is because... well we need
+ * this info later :-)
+ */
+struct enum_cache {
+	unsigned short pref;
+	unsigned short order;
+
+	u_char *tech;
+	int techlen;
+	u_char *dst;
+	int dstlen;
+
+	int shown;
+	char *channel_uniqueid;
+	time_t birth;
+	struct enum_cache *next;
+};
+struct enum_cache *enum_caches = NULL;
 struct enum_context {
 	char *dst;	/* Destination part of URL from ENUM */
 	int dstlen;	/* Length */
@@ -253,6 +286,7 @@
 	char *txt;	/* TXT record in TXT lookup */
 	int txtlen;	/* Length */
 	char *naptrinput;	/* The number to lookup */
+ 	struct ast_channel *chan;
 };
 
 /*--- txt_callback: Callback for TXT record lookup */
@@ -275,7 +309,7 @@
 
 	/* answer is not null-terminated, but should be */
 	/* this is safe to do, as answer has extra bytes on the end we can 
-           safely overwrite with a null */
+	   safely overwrite with a null */
 	answer[len] = '\0';
 	/* now increment len so that len includes the null, so that we can
 	   compare apples to apples */
@@ -290,26 +324,157 @@
 	return 1;
 }
 
+/*--- enum_cache_add: add the data from a NAPTR record to the results for this channel */
+static int enum_cache_add(
+	char *uniqueid,
+	u_char *dst,int dstlen,
+	u_char *tech,int techlen,
+	unsigned short order,unsigned short pref) {
+
+	struct enum_cache *a;
+
+	/* No need to mutex-lock here, it is down outside the call-back
+	   function.
+	*/
+
+	if ((a=(struct enum_cache *)malloc(sizeof(struct enum_cache)))==NULL) {
+		ast_log(LOG_ERROR, "a: malloc returned NULL\n");
+		return -1;
+	}
+	if ((a->dst=(char *)malloc(dstlen))==NULL) {
+		ast_log(LOG_ERROR, "a->dst: malloc returned NULL\n");
+		return -1;
+	}
+	if ((a->tech=(char *)malloc(techlen))==NULL) {
+		ast_log(LOG_ERROR, "a->tech: malloc returned NULL\n");
+		return -1;
+	}
+
+	memcpy(a->dst,dst,dstlen);
+	memcpy(a->tech,tech,techlen);
+	a->dstlen=dstlen;
+	a->techlen=techlen;
+
+	a->pref=pref;
+	a->order=order;
+
+	a->channel_uniqueid=strdup(uniqueid);
+	a->birth=time(NULL);
+	a->shown=0;
+
+	a->next=enum_caches;
+	enum_caches=a;
+
+	return 1;
+}
+
+/*--- enum_cache_cleanup: remove parts of the cache which are older than 5 minutes */
+static int enum_cache_cleanup(void) {
+	struct enum_cache *prev,*this,*next;
+	time_t cutoff=time(NULL)-300;
+	
+	while (enum_caches!=NULL && enum_caches->birth<cutoff) {
+		next=enum_caches->next;
+		free(enum_caches->dst);
+		free(enum_caches->tech);
+		free(enum_caches->channel_uniqueid);
+		free(enum_caches);
+		enum_caches=next;
+	}
+	next=enum_caches;
+	prev=enum_caches; /* This works because the first one doesn't expire */
+	while ((this=next)!=NULL) {
+		next=this->next;
+		if (this->birth<cutoff) {
+			free(this->dst);
+			free(this->tech);
+			free(this->channel_uniqueid);
+			prev->next=next;
+			free(this);
+		} else {
+			prev=this;
+		}
+	}
+	return 0;
+}
+
+/*--- enum_cache_find: find the first not-yet-returned entry for this channel.
+	Returns 0 if there are no more entries */
+static int enum_cache_find(char *uniqueid,struct enum_cache **entry) {
+	struct enum_cache *c,*c_next;
+	struct enum_cache *lowest=NULL;
+	int channelseen=0;
+
+	/* No need to mutex-lock, that is done outside this function */
+
+	/* Find all answers with my uniqueid (which is the channel_uniqueid) */
+	c_next=enum_caches;
+	while ((c=c_next)!=NULL) {
+		c_next=c->next;
+
+		if (strcmp(c->channel_uniqueid,uniqueid)!=0) continue;	/* Not interesting */;
+
+		/* In the DNS answers found earlier, find the first one
+		   already returned. In the mean time, keep track of
+		   the lowerst preference and order found.
+		*/
+
+		channelseen=1;
+
+		if (c->shown!=0) continue;	/* Already done */
+
+/*
+printf("not yet shown: %p - %u %u - %s %s\n",c,c->pref,c->order,c->tech,c->dst);
+*/
+
+		/* First one is always good */
+		if (lowest==NULL) {
+			lowest=c;
+			continue;
+		}
+
+		/* If there is one with a lower order or with the same
+		   order but a lower preference, tag that one.
+		*/
+		if ((c->order<lowest->order)
+		 || (c->order==lowest->order && c->pref<lowest->pref)) {
+			lowest=c;
+			continue;
+		}
+	}
+	*entry=NULL;
+	if (lowest!=NULL) {
+		lowest->shown=1;
+		*entry=lowest;
+	}
+	return channelseen;
+}
+
+
 /*--- enum_callback: Callback from ENUM lookup function */
 static int enum_callback(void *context, u_char *answer, int len, u_char *fullanswer)
 {
 	struct enum_context *c = (struct enum_context *)context;
+	unsigned short order,pref;
 
-	if (parse_naptr(c->dst, c->dstlen, c->tech, c->techlen, answer, len, c->naptrinput)) {
+	if (parse_naptr(c->dst, c->dstlen, c->tech, c->techlen, answer, len, &order, &pref, c->naptrinput)) {
 		ast_log(LOG_WARNING, "Failed to parse naptr :(\n");
 		return -1;
 	}
 
-	if (!ast_strlen_zero(c->dst))
-		return 1;
+	if (!ast_strlen_zero(c->dst)) {
+		if (enum_cache_add(c->chan->uniqueid,c->dst,c->dstlen,c->tech,c->techlen,order,pref)<0)
+			return -1;
+	}
 
 	return 0;
 }
 
-/*--- ast_get_enum: ENUM lookup */
+/*--- ast_get_enum: do the actual lookup for NAPTR records */
 int ast_get_enum(struct ast_channel *chan, const char *number, char *dst, int dstlen, char *tech, int techlen)
 {
 	struct enum_context context;
+	struct enum_cache *c;
 	char tmp[259 + 80];
 	char naptrinput[80] = "+";
 	int pos = strlen(number) - 1;
@@ -325,6 +490,7 @@
 	context.dstlen = dstlen;
 	context.tech = tech;			/* Return string */
 	context.techlen = techlen;
+	context.chan = chan;
 
 	if (pos > 128)
 		pos = 128;
@@ -336,29 +502,44 @@
 	if (chan && ast_autoservice_start(chan) < 0)
 		return -1;
 
-	for(;;) {
-		ast_mutex_lock(&enumlock);
-		if (version != enumver) {
-			/* Ooh, a reload... */
-			s = toplevs;
-			version = enumver;
-		} else {
-			s = s->next;
+	ast_mutex_lock(&enumlock);
+	if (version != enumver) {
+		/* Ooh, a reload... */
+		s = toplevs;
+		version = enumver;
+	} else {
+		s = s->next;
+	}
+	if (s) {
+		strncpy(tmp + newpos, s->toplev, sizeof(tmp) - newpos - 1);
+	}
+	ast_mutex_unlock(&enumlock);
+	if (!s)
+		return 0;
+
+	/* ... to here is black magic for me */
+
+	ret=0;
+	ast_mutex_lock(&enumcachelock);
+	if (enum_cache_find(chan->uniqueid,&c)==0) {
+		enum_cache_cleanup();
+		ret = ast_search_dns(&context, tmp, C_IN, T_NAPTR, enum_callback);
+		if (ret < 0) {
+			ast_log(LOG_DEBUG, "No such number found: %s (%s)\n", tmp, strerror(errno));
+			ret = 0;
 		}
-		if (s) {
-			strncpy(tmp + newpos, s->toplev, sizeof(tmp) - newpos - 1);
+		if (enum_cache_find(chan->uniqueid,&c)==0) {
+			ast_log(LOG_DEBUG, "No such pid found: %s (%s)\n", tmp, strerror(errno));
+			ret = 0;
 		}
-		ast_mutex_unlock(&enumlock);
-		if (!s)
-			break;
-		ret = ast_search_dns(&context, tmp, C_IN, T_NAPTR, enum_callback);
-		if (ret > 0)
-			break;
-	}
-	if (ret < 0) {
-		ast_log(LOG_DEBUG, "No such number found: %s (%s)\n", tmp, strerror(errno));
-		ret = 0;
 	}
+	if (c!=NULL) {
+		memcpy(dst,c->dst,dstlen);
+		memcpy(tech,c->tech,techlen);
+		ret=1;
+	} 
+	ast_mutex_unlock(&enumcachelock);
+
 	if (chan)
 		ret |= ast_autoservice_stop(chan);
 	return ret;
@@ -370,6 +551,7 @@
 int ast_get_txt(struct ast_channel *chan, const char *number, char *dst, int dstlen, char *tech, int techlen, char *txt, int txtlen)
 {
 	struct enum_context context;
+	struct enum_cache *c;
 	char tmp[259 + 80];
 	char naptrinput[80] = "+";
 	int pos = strlen(number) - 1;
@@ -378,6 +560,7 @@
 	struct enum_search *s = NULL;
 	int version = -1;
 
+	/* From here.... */
 	strncat(naptrinput, number, sizeof(naptrinput) - 2);
 
 	context.naptrinput = naptrinput;
@@ -387,6 +570,7 @@
 	context.techlen = techlen;
 	context.txt = txt;
 	context.txtlen = txtlen;
+	context.chan = chan;
 
 	if (pos > 128)
 		pos = 128;
@@ -398,30 +582,44 @@
 	if (chan && ast_autoservice_start(chan) < 0)
 		return -1;
 
-	for(;;) {
-		ast_mutex_lock(&enumlock);
-		if (version != enumver) {
-			/* Ooh, a reload... */
-			s = toplevs;
-			version = enumver;
-		} else {
-			s = s->next;
+	ast_mutex_lock(&enumlock);
+	if (version != enumver) {
+		/* Ooh, a reload... */
+		s = toplevs;
+		version = enumver;
+	} else {
+		s = s->next;
+	}
+	if (s) {
+		strncpy(tmp + newpos, s->toplev, sizeof(tmp) - newpos - 1);
+	}
+	ast_mutex_unlock(&enumlock);
+	if (!s)
+		return 0;
+
+	/* ... to here is black magic for me */
+
+	ret=0;
+	ast_mutex_lock(&enumcachelock);
+	if (enum_cache_find(chan->uniqueid,&c)==0) {
+		enum_cache_cleanup();
+		ret = ast_search_dns(&context, tmp, C_IN, T_NAPTR, enum_callback);
+		if (ret < 0) {
+			ast_log(LOG_DEBUG, "No such number found: %s (%s)\n", tmp, strerror(errno));
+			ret = 0;
 		}
-		if (s) {
-			strncpy(tmp + newpos, s->toplev, sizeof(tmp) - newpos - 1);
+		if (enum_cache_find(chan->uniqueid,&c)==0) {
+			ast_log(LOG_DEBUG, "No such channel->uniqueid found: %s (%s)\n", chan->uniqueid, strerror(errno));
+			ret = 0;
 		}
-		ast_mutex_unlock(&enumlock);
-		if (!s)
-			break;
-
-		ret = ast_search_dns(&context, tmp, C_IN, T_TXT, txt_callback);
-		if (ret > 0)
-			break;
-	}
-	if (ret < 0) {
-		ast_log(LOG_DEBUG, "No such number found: %s (%s)\n", tmp, strerror(errno));
-		ret = 0;
 	}
+	if (c!=NULL) {
+		memcpy(dst,c->dst,dstlen);
+		memcpy(tech,c->tech,techlen);
+		ret=1;
+	} 
+	ast_mutex_unlock(&enumcachelock);
+
 	if (chan)
 		ret |= ast_autoservice_stop(chan);
 	return ret;
